import asyncio
from aiohttp import web
from pandas.core.frame import DataFrame
import numpy as np
import torch
import time
from datetime import datetime
import pandas as pd
import json
from pathlib import Path
import os
from .model import LSTM, model_path
from . import data

data_path = Path(os.path.abspath(os.path.dirname(__file__))) / ".." / "data"

TRAFFIC_INTERVAL = 0.05  # second


def prepare_data():
    traffic_data: DataFrame = data.load_data("KDDTest+.txt")
    traffic_data_input, _, _ = data.prepare_data(traffic_data, True)

    network_data: DataFrame = pd.read_csv(
        data_path / "network.txt",
        header=None,
        comment="#",
        sep="\t",
        names=["FromRowId", "ToRowId"],
    ).iloc[:300]

    network_src_nodes = network_data.FromRowId.unique()
    network_dst_nodes = network_data.ToRowId.unique()
    all_nodes = np.unique(np.concatenate((network_src_nodes, network_dst_nodes)))

    # randomly assign src and dst to traffics
    # random_src = pd.Series(
    #     network_src_nodes[
    #         (np.random.rand(len(traffic_data.index)) * len(network_src_nodes)).astype(
    #             int
    #         )
    #     ]
    # )

    # random_dst = pd.Series(
    #     network_dst_nodes[
    #         (np.random.rand(len(traffic_data.index)) * len(network_dst_nodes)).astype(
    #             int
    #         )
    #     ]
    # )

    random_edge = network_data.iloc[
        (np.random.rand(len(traffic_data.index)) * len(network_data.index)).astype(int)
    ]

    traffic_data["src"] = pd.Series(random_edge.FromRowId.array)
    traffic_data["dst"] = pd.Series(random_edge.ToRowId.array)

    network_json = [
        {"group": "nodes", "data": {"id": str(node)}} for node in all_nodes
    ] + [
        {
            "group": "edges",
            "data": {
                "id": f"{src}-{dst}",
                "source": str(src),
                "target": str(dst),
            },
        }
        for (src, dst) in network_data.itertuples(index=False)
    ]

    return traffic_data, traffic_data_input, network_json


def load_model():
    model = LSTM()
    model = model.double()
    model.load_state_dict(torch.load(model_path / f"{model.__class__.__name__}.pt"))

    return model


def start():
    traffic_data, traffic_data_input, network_data = prepare_data()
    model = load_model()

    print(traffic_data_input.shape)

    is_paused = None

    async def get_network(req):
        return web.json_response(network_data)

    async def get_traffic(req):
        nonlocal is_paused

        ws = web.WebSocketResponse()
        await ws.prepare(req)

        # reset it
        is_paused = asyncio.get_event_loop().create_future()

        with torch.no_grad():
            while True:
                iter = traffic_data.iterrows()
                for index, row in iter:

                    if is_paused is not None:
                        await is_paused

                    prediction = model(traffic_data_input[index].unsqueeze(0)).squeeze()
                    likelihood = torch.nn.functional.softmax(prediction, 0)

                    top_n, top_i = likelihood.topk(1)
                    category_i = top_i[0].item()
                    confidence = top_n[0].item()

                    row = row.to_dict()

                    if category_i == 0:
                        row["prediction"] = {
                            "class": "normal",
                            "confidence": confidence,
                        }
                    else:
                        row["prediction"] = {
                            "class": "abnormal",
                            "confidence": confidence,
                        }

                    # add in the timestamp
                    row["timestamp"] = datetime.now().isoformat()
                    await ws.send_str(json.dumps(row))
                    await asyncio.sleep(np.random.random() * 0.1 + TRAFFIC_INTERVAL)

    async def post_pause(req):
        nonlocal is_paused

        if is_paused is None or is_paused.done():
            is_paused = asyncio.get_event_loop().create_future()

        return web.Response(text="")

    async def post_resume(req):
        nonlocal is_paused
        if is_paused is not None:
            is_paused.set_result(None)
            is_paused = None

        return web.Response(text="")

    app = web.Application()
    app.add_routes(
        [
            web.get("/network", get_network),
            web.get("/traffic", get_traffic),
            web.post("/control/pause", post_pause),
            web.post("/control/resume", post_resume),
        ]
    )

    web.run_app(app, port=3010)
