import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from os import device_encoding, path
from pathlib import Path
import torch
from torch.nn import CrossEntropyLoss
from torch.optim import Adam, Adagrad, Optimizer
from torch.utils.data import TensorDataset, DataLoader
import time

from .model import LSTM, CNN, model_path
from .data import load_data, prepare_data

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

workspace_path = Path(path.abspath(path.dirname(__file__))) / ".."
data_path = workspace_path / "data"

BATCH_SIZE = 32


def accuracy(preds, y):
    """
    Returns accuracy per batch, i.e. if you get 8/10 right, this returns 0.8, NOT 8
    """
    predictedLabel = torch.argmax(torch.exp(preds), dim=1)
    correct = (predictedLabel == y).float()  # convert into float for division
    acc = correct.sum() / len(correct)
    return acc


def get_model(model_type, *args, **argv):
    if model_type == "lstm":
        net = LSTM(*args, **argv)
    else:
        net = CNN(*args, **argv)

    net = net.double()

    net.to(device)

    optimizer = Adam(net.parameters())

    return (net, optimizer)


def train(model, batch, optimizer: Optimizer, criterion):
    epoch_loss = 0
    epoch_acc = 0

    model.train()

    for d, label, _ in batch:
        optimizer.zero_grad()
        predictions = model(d).squeeze(1)

        loss = criterion(predictions, label)

        acc = accuracy(predictions, label)

        loss.backward()

        optimizer.step()

        epoch_loss += loss.item()
        epoch_acc += acc.item()

    return epoch_loss / len(batch), epoch_acc / len(batch)


def evaluate(model, iterator, criterion):
    epoch_loss = 0
    epoch_acc = 0

    model.eval()

    with torch.no_grad():
        for d, label, _ in iterator:
            predictions = model(d).squeeze(1)

            loss = criterion(predictions, label)

            acc = accuracy(predictions, label)

            epoch_loss += loss.item()
            epoch_acc += acc.item()

    return epoch_loss / len(iterator), epoch_acc / len(iterator)


def epoch_time(start_time, end_time):
    elapsed_time = end_time - start_time
    elapsed_mins = int(elapsed_time / 60)
    elapsed_secs = int(elapsed_time - (elapsed_mins * 60))
    return elapsed_mins, elapsed_secs


def prepare_iterator(name, sequence):
    x, lables, levels = prepare_data(load_data(name), sequence=sequence)
    ds = TensorDataset(
        x.to(device),
        lables.to(device),
        levels.to(device),
    )
    dl = DataLoader(ds, BATCH_SIZE)

    return dl


def train_loop(model_type="lstm"):
    sequence = model_type == "lstm"

    train_data = prepare_iterator("KDDTrain+.txt", sequence)
    val_data = prepare_iterator("KDDTest+.txt", sequence)

    model, optimizer = get_model(model_type, output_dim=2)

    loss = CrossEntropyLoss()
    loss = loss.to(device)

    train_losses = []
    val_losses = []

    epoches = range(15)
    best_accuracy = 0

    print("Start training...")

    for epoch in epoches:
        start_time = time.time()

        train_loss, train_acc = train(model, train_data, optimizer, loss)
        val_loss, val_acc = evaluate(model, val_data, loss)

        end_time = time.time()

        epoch_mins, epoch_secs = epoch_time(start_time, end_time)

        train_losses.append(train_acc)
        val_losses.append(val_acc)

        print(f"Epoch: {epoch+1:02} | {epoch_mins}m {epoch_secs}s")
        print(f"\tTrain Loss: {train_loss:.3f} | Train Acc: {train_acc*100:.2f}%")
        print(f"\tVal Loss: {val_loss:.3f} | Val Acc: {val_acc*100:.2f}%")

        if val_acc > best_accuracy:
            best_accuracy = val_acc
            print(f"\tSaving this model...")
            torch.save(
                model.state_dict(), model_path / f"{model.__class__.__name__}.pt"
            )

    plt.plot(epoches, train_losses, label="train")
    plt.plot(epoches, val_losses, label="val")
    plt.legend()
    plt.show()
